from ...repos import LabelMappingsDBRepo
from ...models import LabelMapping

def retrieve_campaign_label_mappings(campaign_id):
	label_mappings_repo = LabelMappingsDBRepo()
	data = label_mappings_repo.retrieve_label_mappings(campaign_id)
	label_mappings = [LabelMapping.from_tuple(row) for row in data]
	return label_mappings



from ...repos import CampaignsDBRepo, ImagesDBRepo
from ...models import Campaign, Image

def increment_annotation_counter(image_id, value = 1):
	campaign_repo = CampaignsDBRepo()
	image_repo = ImagesDBRepo()
	data = image_repo.retrieve_image(image_id)
	if not data:
			raise Exception("Image not found")
	image = Image.from_tuple(data)
	data = campaign_repo.retrieve_one_campaign(image.campaign_id)
	if not data:
		raise Exception("Campaign not found")
	campaign = Campaign.from_tuple(data)
	campaign.annotation_count += value
	campaign_repo.update_campaign(campaign)
	return campaign


import rasterio as rio
import numpy as np
import PIL
import requests
from io import BytesIO
from rasterio import features
import os

from ...repos import ModelsDBRepo, ImagesDBRepo, CampaignsDBRepo, EOTDLRepo, LabelMappingsDBRepo, LabelsDBRepo
from ...models import Image, Model, Campaign, LabelMapping, Label
from . import processing
from ..annotations import create_segmentation_annotation

def sigmoid(x):
	return 1 / (1 + np.exp(-x))

def parse_processing_step(step):
    if isinstance(step, str):
        name = step.split("(")[0]
        # params = step.split("(")[1].split(")")[0].split(", ")
        # kwargs = {param.split("=")[0]: param.split("=")[1] for param in params}
        # return getattr(processing, name)(**kwargs)
        return getattr(processing, name)()
    raise ValueError(f"Invalid processing step: {step}")

def inference_model(model_id: str, image: str):
	# retrieve model
	models_repo = ModelsDBRepo()
	model = models_repo.retrieve_model(model_id)
	model = Model.from_tuple(model)
	# retrieve image
	images_repo = ImagesDBRepo()
	image = images_repo.retrieve_image(image)
	image = Image.from_tuple(image)
	# retrieve campaign
	campaign_repo = CampaignsDBRepo()
	campaign = campaign_repo.retrieve_campaign(image.campaign_id)
	campaign = Campaign.from_tuple(campaign)
	# retrieve label mappings
	label_mappings_repo = LabelMappingsDBRepo()
	label_mappings = label_mappings_repo.retrieve_label_mapping_model(campaign.id, model.id)
	label_mappings = [LabelMapping.from_tuple(d) for d in label_mappings]
	# generate image path 
	if campaign.eotdlDatasetId:
		eotdl_repo = EOTDLRepo()
		image_path = eotdl_repo.get_url(campaign.eotdlDatasetId, image.path)
	else:
		image_path = image.path
	# read image 
	ds = rio.open(image_path)
	transform = ds.transform
	crs = ds.crs
	x = ds.read()
	# apply preprocessing steps
	for step in model.preprocessing:
		x = parse_processing_step(step)(x)
	# save image to memory buffer
	img_buffer = BytesIO()
	# Create memory tif with same metadata as input
	with rio.open(img_buffer, 'w', driver='GTiff',
				height=x.shape[1], width=x.shape[2],
				count=x.shape[0], dtype=x.dtype,
				crs=crs, transform=transform) as dst:
		for i in range(x.shape[0]):
			dst.write(x[i], i+1)
	img_buffer.seek(0)
	# send request with memory buffer
	res = requests.post(model.url, files={'image': (img_buffer)})
	# res = requests.post(model.url, files={'image': ('image.tif', img_buffer, 'image/tiff')})
	if res.status_code != 200:
		raise Exception(f"Error in inference: {res.status_code} {res.text}")
	# generate annotations
	annotations = []
	if model.task == "segmentation":
		# decode the image from the response content
		with rio.open(BytesIO(res.content)) as src:
			# y = src.read(1)  # Read the first band
			# read all bands
			y = src.read()
		print("y", y.shape)
		# output indexes
		output_indexes = [l.output_index for l in label_mappings]
		if max(output_indexes) > y.shape[0]:
			raise ValueError(f"Found output index in label mapping that is larger than the number of bands!")
		# apply postprocessing steps
		for step in model.postprocessing:
			y = parse_processing_step(step)(y)
		for ix, band in enumerate(y):
			# get label from label mapping
			label_id = [l for l in label_mappings if l.output_index == ix]
			if len(label_id) == 0:
				continue # no label mapping for this band
			label_id = label_id[0]
			label_repo = LabelsDBRepo()
			label = label_repo.retrieve_label(label_id)
			label = Label.from_tuple(label)
			# Convert binary mask to vector features
			shapes = features.shapes(
				band.astype(np.uint8),
				transform=transform
			)
			# Convert shapes to GeoJSON features
			# TODO: make compatible with segmentation annotation format
			# {'type': 'Feature', 'properties': {}, 'geometry': {'type': 'MultiPolygon', 'coordinates': [[[[2.805799538073053, 42.35803640805155], ..., [2.805799538073053, 42.35803640805155]]]]}}
			geojson = {
				"type": "FeatureCollection",
				"features": [
					{
						"type": "Feature",
						"geometry": geometry,
						"properties": {"value": value}
					}
					for geometry, value in shapes
				]
			}
			# Add CRS information if available
			if crs:
				geojson["crs"] = {
					"type": "name",
					"properties": {"name": crs.to_string()}
				}
			# save annotation 
			# create_segmentation_annotation(image.id, geojson, label.name) # se guardan los names en annotations o los ids?
		return annotations
	else:
		raise ValueError(f"Not implemented for task {model.task}")


from ...models import Campaign
from ...repos import CampaignsDBRepo

def retrieve_one_campaign(campaign_id):
    repo = CampaignsDBRepo()
    data = repo.retrieve_one_campaign(campaign_id)
    return Campaign.from_tuple(data)


from ...models import Campaign
from ...repos import CampaignsDBRepo, ImagesDBRepo, EOTDLRepo
from .create_campaign import get_bbox
from ...usecases.labels import create_label


async def create_campaign_eotdl(name, description, eotdlDatasetId, labels, labelMappings, progress_callback=None):	
	repo = CampaignsDBRepo()
	id = repo.generate_id()
	campaign = Campaign(id=id, name=name, description=description, eotdlDatasetId=eotdlDatasetId)
	repo.create_campaign(campaign)
	# create labels
	if labels:
		for label in labels:
			create_label(label['name'], label['color'], campaign.id)
	# retrieve images from eotdl
	eotdl_repo = EOTDLRepo()
	images, error = eotdl_repo.get_files(eotdlDatasetId, "*.tif") 
	if len(images) == 0:
		raise Exception("No '.tif' images found in EOTDL dataset.")
	# generate bounding boxes
	if progress_callback is not None:
		bbs = []
		for i, image in enumerate(images):
			await progress_callback((i+1)/len(images), f"Processing {image}")
			bbs.append(get_bbox(eotdl_repo.get_url(eotdlDatasetId, image)))
	else:
		bbs = [get_bbox(eotdl_repo.get_url(eotdlDatasetId, image)) for image in images]
	# store images and bounding boxes
	images_repo = ImagesDBRepo()
	images_repo.create_images(images, id, bbs)
	campaign.image_count = len(images)
	repo.update_campaign(campaign)
	return campaign


from ...models import Annotation
from ...repos import AnnotationsDBRepo

def retrieve_annotations(image_id):
	repo = AnnotationsDBRepo()
	data = repo.retrieve_annotations(image_id)
	return [Annotation.from_tuple(row) for row in data]




from ...repos import PluginsDBRepo
from ...models import Plugin

def disable_plugin(name):
	repo = PluginsDBRepo()
	data = repo.retrieve_plugin(name)
	plugin = Plugin.from_tuple(data)
	plugin.enabled = False
	repo.update_plugin(plugin)
	return plugin


from ...repos import AnnotationsDBRepo
from ...models import Annotation

from .increment_annotation_counter import increment_annotation_counter

def create_detection_annotation(image_id, bb, label):	
	repo = AnnotationsDBRepo()
	annotation = Annotation(id=repo.generate_id(), type="detection", value=label, bb=bb, image_id=image_id)
	repo.create_annotation(annotation)
	increment_annotation_counter(image_id)
	return annotation



from .CampaignsDBRepo import CampaignsDBRepo
from .ImagesDBRepo import ImagesDBRepo
from .ModelsDBRepo import ModelsDBRepo
from .PluginsDBRepo import PluginsDBRepo
from .LabelsDBRepo import LabelsDBRepo
from .AnnotationsDBRepo import AnnotationsDBRepo
from .EOTDLRepo import EOTDLRepo
from .LabelMappingsDBRepo import LabelMappingsDBRepo


from .DBRepo import DBRepo

class LabelMappingsDBRepo(DBRepo):
    def __init__(self):
        super().__init__()
        cursor = self.get_cursor()
        cursor.execute(f"""CREATE TABLE IF NOT EXISTS label_mappings (
            id TEXT PRIMARY KEY,
            campaignId TEXT NOT NULL,
            modelId TEXT NOT NULL,
            labelId TEXT NOT NULL,
            output_index INTEGER NOT NULL,
            FOREIGN KEY (campaignId) REFERENCES campaigns(id),
            FOREIGN KEY (modelId) REFERENCES models(id),
            FOREIGN KEY (labelId) REFERENCES labels(id)
        )""")
        self.commit_and_close_db()

    def label_mapping_exists(self, campaign_id, model_id, label_id):
        cursor = self.get_cursor()
        cursor.execute("SELECT * FROM label_mappings WHERE campaignId = ? AND modelId = ? AND labelId = ?", (campaign_id, model_id, label_id))
        return cursor.fetchone() is not None

    def create_label_mapping(self, label_mapping):
        cursor = self.get_cursor()
        cursor.execute("INSERT INTO label_mappings (id, campaignId, modelId, labelId, output_index) VALUES (?, ?, ?, ?, ?)", 
                       (label_mapping.id, label_mapping.campaignId, label_mapping.modelId, label_mapping.labelId, label_mapping.output_index))
        self.commit_and_close_db()

    def retrieve_label_mappings(self, campaign_id):
        cursor = self.get_cursor()
        cursor.execute("SELECT * FROM label_mappings WHERE campaignId = ?", (campaign_id,))
        return cursor.fetchall()
    
    def retrieve_label_mapping_model(self, campaign_id, model_id):
        cursor = self.get_cursor()
        cursor.execute("SELECT * FROM label_mappings WHERE campaignId = ? AND modelId = ?", (campaign_id, model_id))
        return cursor.fetchall()
    
    def retrieve_label_mappings_model(self, model_id):
        cursor = self.get_cursor()
        cursor.execute("SELECT * FROM label_mappings WHERE modelId = ?", (model_id,))
        return cursor.fetchall()
    
    def delete_label_mappings(self, campaign_id):
        cursor = self.get_cursor()
        cursor.execute("DELETE FROM label_mappings WHERE campaignId = ?", (campaign_id,))
        self.commit_and_close_db()

    def delete_label_mapping_model(self, model_id):
        cursor = self.get_cursor()
        cursor.execute("DELETE FROM label_mappings WHERE modelId = ?", (model_id,))
        self.commit_and_close_db()
    
    def delete_label_mapping_label(self, campaign_id, model_id, label_id):
        cursor = self.get_cursor()
        cursor.execute("DELETE FROM label_mappings WHERE campaignId = ? AND modelId = ? AND labelId = ?", (campaign_id, model_id, label_id))
        self.commit_and_close_db()

    def update_label_mapping(self, label_mapping):
        cursor = self.get_cursor()
        cursor.execute("UPDATE label_mappings SET output_index = ? WHERE campaignId = ? AND modelId = ? AND labelId = ?", 
                       (label_mapping.output_index, label_mapping.campaignId, label_mapping.modelId, label_mapping.labelId))
        self.commit_and_close_db()

    def delete_label_mapping(self, id):
        cursor = self.get_cursor()
        cursor.execute("DELETE FROM label_mappings WHERE id = ?", (id,))
        self.commit_and_close_db()


from ...repos import ModelsDBRepo, LabelMappingsDBRepo

def delete_model(id):
	repo = ModelsDBRepo()
	repo.delete_model(id)
	label_mappings_repo = LabelMappingsDBRepo()
	label_mappings_repo.delete_label_mapping_model(id)
	return {"message": "Model deleted"}



import requests
from io import BytesIO
import fnmatch
import os
import json

class EOTDLRepo:
    def __init__(self):
        self.url = "https://api.eotdl.com/"
        # get user credentials from env or creds file
        self.user = None 
        if 'EOTDL_API_KEY' in os.environ:
            self.user = {'api_key': os.environ['EOTDL_API_KEY']}
        else:
            try:
                with open(os.path.expanduser('~/.eotdl/creds.json')) as f:
                    self.user = json.load(f)
            except FileNotFoundError:
                raise Exception("No EOTDL credentials found. Please run `eotdl auth login` or set the EOTDL_API_KEY environment variable.")

    def format_response(self, response):
        if response.status_code == 200:
            return response.json(), None
        return None, response.json()["detail"]

    def generate_headers(self, data):
        if "api_key" in data:
            return {"X-API-Key": data["api_key"]}
        if "id_token" in data:
            return {"Authorization": "Bearer " + data["id_token"]}
        raise Exception("Invalid headers")

    def get_url(self, dataset_id, filename, version=None):
        url = self.url + f"datasets/{dataset_id}/download/{filename}"
        if version is not None:
            url += "?version=" + str(version)
        headers = self.generate_headers(self.user)
        response = requests.get(url, headers=headers, stream=True)
        return BytesIO(response.content)
    
    def get_files(self, dataset_id, pattern=None):
        res = requests.get(f"https://api.eotdl.com/datasets/{dataset_id}/files")
        files, error = self.format_response(res)
        if error:
            return None, error
        files = [f["filename"] for f in files]
        if pattern:
            return fnmatch.filter(files, pattern), None
        return files, None
    


from fastapi import APIRouter, HTTPException, status
from fastapi.responses import StreamingResponse
from pydantic import BaseModel

from src.usecases.images import retrieve_images
from src.utils.image import get_image_data, get_tile_data, ready_image
from src.utils.image.errors import ImageOutOfBounds
from src.repos import EOTDLRepo
router = APIRouter(prefix="/images", tags=["images"])

@router.get("/{campaign}")
def _retrieve_images(campaign: str):
    try:
        return retrieve_images(campaign)
    except Exception as e:
        print("error images:retrieve_images", e)
        raise HTTPException(status_code=500, detail=str(e))

@router.get("/{image:path}/{z}/{x}/{y}.png")
def retrieve_image_tile(
    image: str,
    z: int,
    x: int,
    y: int,
    bands: str = "4,3,2",
    stretch: str = "0,3000",
    palette: str = "viridis",
    eotdlDatasetId: str = None
):
    tile_size = (256, 256)
    if len(bands) == 1:
        bands = int(bands)
    else:
        bands = tuple([int(band) for band in bands.split(",")])
    stretch = tuple([float(v) for v in stretch.split(",")])
    try:
        if eotdlDatasetId:
            eotdl_repo = EOTDLRepo()
            image = eotdl_repo.get_url(eotdlDatasetId, image)
        tile = get_tile_data(image, (x, y, z), bands, tile_size)
        tile = get_image_data(tile, stretch, palette)
        image = ready_image(tile)
        return StreamingResponse(image, media_type="image/png")
    except ImageOutOfBounds as error:
        print(error)
        raise HTTPException(status_code=status.HTTP_404_NOT_FOUND, detail=error.message)
    except Exception as e:
        print("error images:retrieve_image_tile", e)
        raise HTTPException(status_code=500, detail="Could not retrieve tile")



from ...repos import AnnotationsDBRepo
from ...models import Annotation

from .increment_annotation_counter import increment_annotation_counter

def create_classification_annotation(image_id, label):	
	repo = AnnotationsDBRepo()
	data = repo.get_annotation_by_image_id(image_id, label, type="classification")
	if not data:
		print("no annotation found")
		annotation = Annotation(id=repo.generate_id(), type="classification", value=label, image_id=image_id)
		repo.create_annotation(annotation)
		increment_annotation_counter(image_id)
		return annotation
	print("annotation found")
	annotation = Annotation.from_tuple(data)
	annotation.type = "classification" 
	annotation.value = label 
	repo.update_annotation(annotation)
	return annotation



from .retrieve_labels import retrieve_labels
from .create_label import create_label
from .delete_label import delete_label
from .generate_labels_json import generate_labels_json
from .save_labels import save_labels



from pydantic import BaseModel, Field
from datetime import datetime
from typing import Optional

class Label(BaseModel):
    id: str
    name: str
    campaign_id: str
    color: str
    createdAt: datetime = Field(default_factory=datetime.now)
    updatedAt: datetime = Field(default_factory=datetime.now)

    @classmethod
    def from_tuple(cls, data: tuple):
        return cls(
            id=data[0],
            name=data[1],
            color=data[2],
            campaign_id=data[3],
            createdAt=datetime.fromisoformat(data[4]),
            updatedAt=datetime.fromisoformat(data[5])
        )



from ...repos import AnnotationsDBRepo
from ...models import Annotation

from .increment_annotation_counter import increment_annotation_counter

def delete_annotation(id):
	repo = AnnotationsDBRepo()
	data = repo.retrieve_one_annotation(id)
	if not data:
		raise Exception("Annotation not found")
	annotation = Annotation.from_tuple(data)
	repo.delete_annotation(id)
	increment_annotation_counter(annotation.image_id, -1)
	return {"message": "Annotation deleted"}


from ...models import Model, LabelMapping, Model
from ...repos import ModelsDBRepo, LabelMappingsDBRepo

def retrieve_models(campaign_id: str = None):
    repo = ModelsDBRepo()
    if not campaign_id:
        data = repo.retrieve_models()
        models = [Model.from_tuple(d) for d in data]
        return models
    # retrieve models that have label mapping to the campaign
    repo = ModelsDBRepo()
    label_mappings_repo = LabelMappingsDBRepo()
    label_mappings = label_mappings_repo.retrieve_label_mappings(campaign_id) 
    label_mappings = [LabelMapping.from_tuple(d) for d in label_mappings]
    model_ids = [lm.modelId for lm in label_mappings]
    models = repo.retrieve_models_by_ids(model_ids)
    models = [Model.from_tuple(d) for d in models]
    return models




from pydantic import BaseModel, field_validator, Field
from datetime import datetime
from typing import Optional

class Campaign(BaseModel):
    id: str
    name: str
    description: str
    createdAt: datetime = Field(default_factory=datetime.now)
    updatedAt: datetime = Field(default_factory=datetime.now)
    image_count: int = 0
    annotation_count: int = 0
    eotdlDatasetId: Optional[str] = None
    # storage
    # labels

    # TODO: checks and validations
      
    # @field_validator("name")
    # def check_name_is_valid(cls, name):
    #     if name is not None:
    #         assert validate_name(name) == name
    #     return name


    @classmethod
    def from_tuple(cls, data: tuple):
        return cls(
            id=data[0],
            name=data[1],
            description=data[2],
            createdAt=datetime.fromisoformat(data[3]),
            updatedAt=datetime.fromisoformat(data[4]),
            image_count=data[5],
            annotation_count=data[6],
            eotdlDatasetId=data[7]
        )
    
    model_config = {
        "json_encoders": {
            datetime: lambda dt: dt.isoformat()
        }
    }


from ...repos import LabelsDBRepo, LabelMappingsDBRepo

def delete_label(id):
	repo = LabelsDBRepo()
	repo.delete_label(id)
	repo_label_mappings = LabelMappingsDBRepo()
	repo_label_mappings.delete_label_mapping_label(id)
	return {"message": "label deleted"}



from .DBRepo import DBRepo

class ImagesDBRepo(DBRepo):
    def __init__(self):
        super().__init__()
        cursor = self.get_cursor()
        cursor.execute(f"""CREATE TABLE IF NOT EXISTS images (
            id INTEGER PRIMARY KEY AUTOINCREMENT,
            path TEXT NOT NULL,
            campaign_id TEXT,
            bbox TEXT,
            FOREIGN KEY (campaign_id) REFERENCES campaigns(id)
        )""")
        self.commit_and_close_db()
    
    def create_images(self, images, campaign_id, bbs):
        cursor = self.get_cursor()
        cursor.executemany("INSERT INTO images (path, campaign_id, bbox) VALUES (?, ?, ?)", [(image, campaign_id, bb) for image, bb in zip(images, bbs)])
        self.commit_and_close_db()

    def retrieve_images(self, campaign_id):
        cursor = self.get_cursor()
        cursor.execute(f"SELECT * FROM images WHERE campaign_id = ?", (campaign_id,))
        return cursor.fetchall()
    
    def delete_images(self, campaign_id):
        cursor = self.get_cursor()
        cursor.execute(f"DELETE FROM images WHERE campaign_id = ?", (campaign_id,))
        self.commit_and_close_db()

    def retrieve_image(self, image_id):
        cursor = self.get_cursor()
        cursor.execute(f"SELECT * FROM images WHERE id = ?", (image_id,))
        return cursor.fetchone()
    
    def retrieve_image_by_path(self, path, campaign_id):
        cursor = self.get_cursor()
        cursor.execute(f"SELECT * FROM images WHERE path = ? AND campaign_id = ?", (path, campaign_id))
        return cursor.fetchone()


from ...repos import LabelMappingsDBRepo, LabelsDBRepo
from ...models import LabelMapping, Label
from ...usecases.models import create_label_mapping

def update_label_mappings(campaign_id, labelMappings):
	# delete label mappings that are not in the new labelMappings
	repo = LabelMappingsDBRepo()
	current_label_mappings = repo.retrieve_label_mappings(campaign_id)
	current_label_mappings = [LabelMapping.from_tuple(data) for data in current_label_mappings]
	for label_mapping in current_label_mappings:
		if label_mapping.modelId not in labelMappings:
			repo.delete_label_mapping(label_mapping.id)
	# create new label mappings or update existing ones
	label_mappings = []
	for modelId, mapping in labelMappings.items():
		for labelName, index in mapping.items():
			if index:
				data = create_label_mapping(campaign_id, modelId, labelName, index)
				label_mappings.append(data)
			else:
				# delete if existing and not in the new labelMappings
				labels_repo = LabelsDBRepo()
				data = labels_repo.get_label_by_name(labelName, campaign_id)
				label = Label.from_tuple(data)
				repo.delete_label_mapping_label(campaign_id, modelId, label.id)
	return label_mappings



from ...models import Label
from ...repos import LabelsDBRepo
from ...errors.labels import LabelAlreadyExistsError

def create_label(name, color, campaign):	
	repo = LabelsDBRepo()
	if repo.get_label_by_name(name, campaign):
		raise LabelAlreadyExistsError()
	label = Label(id=repo.generate_id(), name=name, campaign_id=campaign, color=color)
	repo.create_label(label)
	return label


from pydantic import BaseModel, Field
from datetime import datetime
from typing import Optional
import json

class Annotation(BaseModel):
    id: str
    type: str
    value: str
    bb: Optional[list[list[float]]] = None
    layer_data: object = None
    image_id: int
    createdAt: datetime = Field(default_factory=datetime.now)
    updatedAt: datetime = Field(default_factory=datetime.now)

    @classmethod
    def from_tuple(cls, data: tuple):
        return cls (
            id=data[0],
            type=data[1],
            value=data[2],
            bb=json.loads(data[3]) if data[3] else None,
            layer_data=json.loads(data[4]) if data[4] else None,
            image_id=data[5],
            createdAt=datetime.fromisoformat(data[6]),
            updatedAt=datetime.fromisoformat(data[7])
        )




from .create_classification_annotation import create_classification_annotation
from .retrieve_annotations import retrieve_annotations
from .delete_annotation import delete_annotation
from .create_detection_annotation import create_detection_annotation
from .update_detection_annotation import update_detection_annotation
from .create_segmentation_annotation import create_segmentation_annotation


from ...models import Annotation 
from ...models import Image

from ...repos import ImagesDBRepo
from .generate_detection_geojson import generate_detection_geojson
from .generate_segmentation_geojson import generate_segmentation_geojson
from .generate_clasification_geojson import generate_clasification_geojson

def generate_geojson(annotations_data: list[list[Annotation]]):
    features = []
    images = []
    images_repo = ImagesDBRepo()
    for image_annotations in annotations_data:
        for annotation in image_annotations:
            if annotation.layer_data:
                features.append(generate_segmentation_geojson(annotation.layer_data, annotation.value))
            elif annotation.bb:
                features.append(generate_detection_geojson(annotation.bb, annotation.value))
            else:
                features.append(generate_clasification_geojson(annotation.value))
        feature_collection = {
            "type": "FeatureCollection",
            "features": features
        }
        if image_annotations:
            image = Image.from_tuple(images_repo.retrieve_image(image_annotations[0].image_id))
            images.append((image.path, feature_collection))
        features = []
    return images



from ...repos import AnnotationsDBRepo
from ...repos import ImagesDBRepo
from ...models import Image
from ...models import Annotation

def get_campaign_annotations(campaign_id):
    annotations_repo = AnnotationsDBRepo()
    images_repo = ImagesDBRepo()
    data = images_repo.retrieve_images(campaign_id)
    images = [Image.from_tuple(d) for d in data]
    annotations = []
    for image in images:
        annotations.append([Annotation.from_tuple(d) for d in annotations_repo.retrieve_annotations(image.id)])
    return annotations




from datetime import datetime
import json

from .DBRepo import DBRepo

class AnnotationsDBRepo(DBRepo):
	def __init__(self):
		super().__init__()
		cursor = self.get_cursor()
		cursor.execute(f"""CREATE TABLE IF NOT EXISTS annotations (
            id TEXT PRIMARY KEY,
			type TEXT NOT NULL,
			value TEXT,
			bb TEXT NULL,
			layer_data TEXT NULL,
			image_id INTEGER,
			createdAt TEXT,
			updatedAt TEXT,
			FOREIGN KEY (image_id) REFERENCES images(id)
		)""")
		self.commit_and_close_db()
		
	def retrieve_annotations(self, image_id):
		cursor = self.get_cursor()
		cursor.execute(f"SELECT * FROM annotations WHERE image_id = ? ORDER BY createdAt DESC", (image_id,))
		return cursor.fetchall()
	
	def retrieve_one_annotation(self, id):
		cursor = self.get_cursor()
		cursor.execute(f"SELECT * FROM annotations WHERE id = ?", (id,))
		return cursor.fetchone()

	def create_annotation(self, annotation):
		cursor = self.get_cursor()
		cursor.execute("INSERT INTO annotations (id, type, value, bb, layer_data, image_id, createdAt, updatedAt) VALUES (?, ?, ?, ?, ?, ?, ?, ?)", (annotation.id, annotation.type, annotation.value, json.dumps(annotation.bb), json.dumps(annotation.layer_data), annotation.image_id, annotation.createdAt, annotation.updatedAt))
		self.commit_and_close_db()

	def update_annotation(self, annotation):
		cursor = self.get_cursor()
		annotation.updatedAt = datetime.now()
		cursor.execute("UPDATE annotations SET type = ?, value = ?, updatedAt = ?, bb = ? WHERE id = ?", (annotation.type, annotation.value, annotation.updatedAt, json.dumps(annotation.bb), annotation.id))
		self.commit_and_close_db()
	
	def delete_annotation(self, id):
		cursor = self.get_cursor()
		cursor.execute(f"DELETE FROM annotations WHERE id = ?", (id,))
		self.commit_and_close_db()

	def get_annotation_by_image_id(self, image_id, value, type):
		cursor = self.get_cursor()
		cursor.execute(f"SELECT * FROM annotations WHERE image_id = ? AND value = ? AND type = ?", (image_id, value, type))
		return cursor.fetchone()



from pydantic import BaseModel, Field
from datetime import datetime
from typing import Optional, List

class Model(BaseModel):
    id: str
    name: str
    description: str
    url: str
    createdAt: datetime = Field(default_factory=datetime.now)
    updatedAt: datetime = Field(default_factory=datetime.now)
    task: str
    preprocessing: Optional[List[str]] = []
    postprocessing: Optional[List[str]] = []

    @classmethod
    def from_tuple(cls, data: tuple):
        return cls(
            id=data[0],
            name=data[1],
            description=data[2],
            url=data[3],
            createdAt=datetime.fromisoformat(data[4]),
            updatedAt=datetime.fromisoformat(data[5]),
            task=data[6],
            preprocessing=data[7].split(',') if data[7] else [],
            postprocessing=data[8].split(',') if data[8] else []
        )
    
class LabelMapping(BaseModel):
    id: str
    campaignId: str
    modelId: str
    labelId: str
    output_index: int

    @classmethod
    def from_tuple(cls, data: tuple):
        return cls(
            id=data[0],
            campaignId=data[1],
            modelId=data[2],
            labelId=data[3],
            output_index=data[4]
        )


from ...models import Annotation
from ...repos import AnnotationsDBRepo
import json
from .get_campaign_annotations import get_campaign_annotations
from .generate_image_geojson import generate_geojson

def save_annotations(campaign_id):
    repo = AnnotationsDBRepo()
    annotations = get_campaign_annotations(campaign_id)
    images_geojson = generate_geojson(annotations)
    for image in images_geojson:
        path = image[0].replace(".tif","_labels.geojson")
        with open(path, "w") as f:
            f.write(json.dumps(image[1]))
    return True



from ...models import Model
from ...repos import ModelsDBRepo

def create_model(name, description, url, task, preprocessing, postprocessing):	
	repo = ModelsDBRepo()
	id = repo.generate_id()
	model = Model(id=id, name=name, description=description, url=url, task=task, preprocessing=preprocessing, postprocessing=postprocessing)
	repo.create_model(model)
	return model


from pydantic import BaseModel

class Image(BaseModel):
    id: int
    path: str
    campaign_id: str
    bbox: str
    # labels

    @classmethod
    def from_tuple(cls, data: tuple):
        return cls(
            id=data[0],
            path=data[1],
            campaign_id=data[2],
            bbox=data[3],
        )


from .retrieve_labels import retrieve_labels

def generate_labels_json(campaign_id):
    campaign_labels = retrieve_labels(campaign_id)
    labels = {
        "labels":[]
    }
    for label in campaign_labels:
        labels["labels"].append({
            "name" : label.name,
            "color" : label.color
        })
    return labels


from ...models import Model
from ...repos import ModelsDBRepo

def retrieve_one_model(model_id):
    repo = ModelsDBRepo()
    data = repo.retrieve_one_model(model_id)
    return Model.from_tuple(data)


from .create_campaign import create_campaign
from .retrieve_campaigns import retrieve_campaigns
from .delete_campaign import delete_campaign
from .retrieve_one_campaign import retrieve_one_campaign
from .create_campaign_eotdl import create_campaign_eotdl
from .create_imported_campaign import create_imported_campaign
from .retrieve_campaign_label_mappings import retrieve_campaign_label_mappings
from .update_label_mappings import update_label_mappings



from ...models import Image
from ...repos import ImagesDBRepo

def retrieve_images(campaign):
    repo = ImagesDBRepo()
    data = repo.retrieve_images(campaign)
    images = [Image.from_tuple(d) for d in data]
    return images


import rasterio as rio
import numpy as np

class ProcessingStep:
    def __init__(self, name, **kwargs):
        self.name = name
        self.parameters = kwargs
        for k,v in kwargs.items():
            setattr(self, k, v)
    def __str__(self):
        params_str = ", ".join([f"{k}={v}" for k,v in self.parameters.items()])
        return f"{self.name} ({params_str})"

class S2RGB(ProcessingStep):
    def __init__(self):
        super().__init__("S2RGB")
    def __call__(self, x):
        x = x[(3,2,1), ...]
        x = np.clip(x / 3000, 0, 1)
        x = (x * 255).astype(np.uint8)
        return x

class Sigmoid(ProcessingStep):
    def __init__(self, threshold=0.5):
        super().__init__("Sigmoid", threshold=threshold)
    def __call__(self, x):
        probas = 1 / (1 + np.exp(-x))
        if self.threshold is not None:
            return probas > self.threshold
        return probas


import typer
import os
import sys
from pathlib import Path

# # Add the cli directory to the Python path
# scaneo_cli_dir = os.path.dirname(os.path.realpath(__file__))
# sys.path.append(os.path.join(scaneo_cli_dir))

app = typer.Typer()

__version__ = "2024.10.09"

@app.command()
def run(
    port: int = typer.Option(8000, "--port", "-p", help="Port to run the server on"),
    reload: bool = typer.Option(
        False, "--reload", "-r", help="Reload the server when files change", hidden=True
    ),
    host: str = typer.Option(
        "localhost", "--host", "-h", help="Host to run the server on"
    ),
    workers: int = typer.Option(
        None, "--workers", "-w", help="Number of workers to run the server on"
    ),
    version: bool = typer.Option(
        False, "--version", "-v", help="Print the version and exit"
    ),
):
    if version:
        typer.echo(f"Scaneo version {__version__}")
        return
    cmd = f"uvicorn api:app --port {port} --host {host} --app-dir {os.path.dirname(os.path.realpath(__file__))}"
    if workers:
        cmd += f" --workers {workers}"
    if reload:
        cmd += " --reload"
    typer.echo(f"Running command: {cmd}")
    os.system(cmd)


if __name__ == "__main__":
    app()



from .DBRepo import DBRepo

class ModelsDBRepo(DBRepo):
    def __init__(self):
        super().__init__()
        cursor = self.get_cursor()
        cursor.execute(f"""CREATE TABLE IF NOT EXISTS models (
            id TEXT PRIMARY KEY,
            name TEXT NOT NULL UNIQUE,
            description TEXT NOT NULL,
            url TEXT NOT NULL,
            createdAt TEXT NOT NULL,
            updatedAt TEXT NOT NULL,
            task TEXT NOT NULL,
            preprocessing TEXT NULL,
            postprocessing TEXT NULL
        )""")
        self.commit_and_close_db()

    def retrieve_models(self):
        cursor = self.get_cursor()
        cursor.execute("SELECT id, name, description, url, createdAt, updatedAt, task, preprocessing, postprocessing FROM models")
        return cursor.fetchall()
    
    def retrieve_models_by_ids(self, ids):
        cursor = self.get_cursor()
        placeholders = ','.join(['?' for _ in ids])
        cursor.execute(f"SELECT id, name, description, url, createdAt, updatedAt, task, preprocessing, postprocessing FROM models WHERE id IN ({placeholders})", ids)
        return cursor.fetchall()
    
    def create_model(self, model):
        cursor = self.get_cursor()
        cursor.execute(f"INSERT INTO models (id, name, description, url, createdAt, updatedAt, task, preprocessing, postprocessing) VALUES (?, ?, ?, ?, ?, ?, ?, ?, ?)", (
            model.id, 
            model.name, 
            model.description, 
            model.url, 
            model.createdAt.isoformat(), 
            model.updatedAt.isoformat(), 
            model.task,
            ','.join(model.preprocessing),
            ','.join(model.postprocessing)
        ))
        self.commit_and_close_db()

    def retrieve_model(self, id):
        cursor = self.get_cursor()
        cursor.execute(f"SELECT * FROM models WHERE id = ?", (id,))
        return cursor.fetchone()

    def delete_model(self, id):
        cursor = self.get_cursor()
        cursor.execute(f"DELETE FROM models WHERE id = ?", (id,))
        self.commit_and_close_db()

    def retrieve_one_model(self, model_id):
        cursor = self.get_cursor()
        cursor.execute("SELECT id, name, description, url, createdAt, updatedAt, task, preprocessing, postprocessing FROM models WHERE id = ?", (model_id,))
        return cursor.fetchone()


from fastapi import APIRouter, HTTPException
from pydantic import BaseModel

from src.usecases.annotations import update_detection_annotation, retrieve_annotations, create_classification_annotation, create_detection_annotation, delete_annotation, create_segmentation_annotation

router = APIRouter(prefix="/annotations", tags=["annotations"])

@router.get("/{image_id}")
def _retrieve_annotations(image_id: str):
    try:
        return retrieve_annotations(image_id)
    except Exception as e:
        print("error annotations:retrieve_annotations", e)
        raise HTTPException(status_code=500, detail=str(e))

class ClassificationBody(BaseModel):
    imageId: int
    label: str

@router.post("/classification")
def _create_classification_annotation(body: ClassificationBody):
    try:
        return create_classification_annotation(body.imageId, body.label)
    except Exception as e:
        print("error annotations:create_classification_annotation", e)
        raise HTTPException(status_code=500, detail=str(e))

class DetectionBody(BaseModel):
    imageId: int
    label: str
    bb: list[list[float]]

@router.post("/detection")
def _create_detection_annotation(body: DetectionBody):
    try:
        return create_detection_annotation(body.imageId, body.bb, body.label)
    except Exception as e:
        print("error annotations:create_detection_annotation", e)
        raise HTTPException(status_code=500, detail=str(e))

@router.delete("/{id}")
def _delete_annotation(id: str):
    try:
        return delete_annotation(id)
    except Exception as e:
        print("error annotations:delete_annotation", e)
        raise HTTPException(status_code=500, detail=str(e))

class UpdateDetectionBody(BaseModel):
    bb: list[list[float]]

@router.put("/detection/{id}")
def _update_detection_annotation(id: str, body: UpdateDetectionBody):
    try:
        return update_detection_annotation(id, body.bb)
    except Exception as e:
        print("error annotations:update_detection_annotation", e)
        return HTTPException(status_code=500, detail=str(e))

class SegmentationBody(BaseModel):
    imageId: int
    label: str
    layer_data: object

@router.post("/segmentation")
def _create_segmentation_annotation(body: SegmentationBody):
    try:
        return create_segmentation_annotation(body.imageId, body.layer_data, body.label)
    except Exception as e:
        print("error annotations:create_segmentation_annotation", e)
        return HTTPException(status_code=500, detail=str(e))


from glob import glob
import os
import rasterio
import json

from ...models import Campaign
from ...repos import CampaignsDBRepo, ImagesDBRepo
from ...usecases.labels import create_label
from ...usecases.models import create_label_mapping

def get_bbox(image):
	src = rasterio.open(image)
	bounds = src.bounds
	crs = src.crs
	if crs != "EPSG:4326":
		bounds = rasterio.warp.transform_bounds(crs, "EPSG:4326", *bounds)
	return json.dumps(bounds)

async def create_campaign(name, description, path, labels, labelMappings, progress_callback=None):	
	repo = CampaignsDBRepo()
	id = repo.generate_id()
	campaign = Campaign(id=id, name=name, description=description)
	repo.create_campaign(campaign)
	# create labels
	if labels:
		for label in labels:
			create_label(label['name'], label['color'], campaign.id)
	# create label mappings
	if labelMappings:
		for modelId, mapping in labelMappings.items():
			for labelName, index in mapping.items():
				create_label_mapping(campaign.id, modelId, labelName, index)
	# retrieve images in path recursively
	images = glob(os.path.join(path, '**/*.tif'), recursive=True)
	images_repo = ImagesDBRepo()
	# generate bounding boxes
	if progress_callback is not None:
		bbs = []
		for i, image in enumerate(images):
			await progress_callback((i+1)/len(images), f"Processing {image}")
			bbs.append(get_bbox(image))
	else:
		bbs = [get_bbox(image) for image in images]
	images_repo.create_images(images, id, bbs)
	campaign.image_count = len(images)
	repo.update_campaign(campaign)
	return campaign


from ...models import Campaign
from ...repos import CampaignsDBRepo, ImagesDBRepo, LabelsDBRepo, LabelMappingsDBRepo

def delete_campaign(id):
	repo = CampaignsDBRepo()
	repo.delete_campaign(id)
	image_repo = ImagesDBRepo()
	image_repo.delete_images(id)
	label_repo = LabelsDBRepo()
	label_repo.delete_labels(id)
	label_mappings_repo = LabelMappingsDBRepo()
	label_mappings_repo.delete_label_mappings(id)
	return {"message": "Campaign deleted"}



from .DBRepo import DBRepo

class LabelsDBRepo(DBRepo):
	def __init__(self):
		super().__init__()
		cursor = self.get_cursor()
		cursor.execute(f"""CREATE TABLE IF NOT EXISTS labels (
            id TEXT PRIMARY KEY,
			name TEXT NOT NULL,
			color TEXT NOT NULL,
			campaign_id TEXT,
			createdAt TEXT,
			updatedAt TEXT,
			FOREIGN KEY (campaign_id) REFERENCES campaigns(id),
			UNIQUE(name, campaign_id)
		)""")
		self.commit_and_close_db()
		
	def retrieve_labels(self, campaign_id):
		cursor = self.get_cursor()
		cursor.execute(f"SELECT * FROM labels WHERE campaign_id = ? ORDER BY createdAt DESC", (campaign_id,))
		return cursor.fetchall()

	def create_label(self, label):
		try:
			cursor = self.get_cursor()
			cursor.execute("INSERT INTO labels (id, name, campaign_id, color, createdAt, updatedAt) VALUES (?, ?, ?, ?, ?, ?)", 
						  (label.id, label.name, label.campaign_id, label.color, label.createdAt, label.updatedAt))
			self.commit_and_close_db()
		except self.sqlite3.IntegrityError as e:
			if "UNIQUE constraint failed" in str(e):
				raise ValueError(f"A label with name '{label.name}' already exists in this campaign")
			raise e

	def delete_label(self, id):
		cursor = self.get_cursor()
		cursor.execute(f"DELETE FROM labels WHERE id = ?", (id,))
		self.commit_and_close_db()

	def get_label_by_name(self, name, campaign_id):
		cursor = self.get_cursor()
		cursor.execute(f"SELECT * FROM labels WHERE name = ? AND campaign_id = ?", (name, campaign_id))
		return cursor.fetchone()
	
	def delete_labels(self, campaign_id):
		cursor = self.get_cursor()
		cursor.execute(f"DELETE FROM labels WHERE campaign_id = ?", (campaign_id,))
		self.commit_and_close_db()


from fastapi import APIRouter, HTTPException
from pydantic import BaseModel

from src.usecases.plugins import retrieve_plugins, enable_plugin, disable_plugin

router = APIRouter(prefix="/plugins", tags=["plugins"])

@router.get("")
def _retrieve_plugins():
    try:
        return retrieve_plugins()
    except Exception as e:
        print("error plugins:retrieve_plugins", e)
        raise HTTPException(status_code=500, detail=str(e))


class Body(BaseModel):
    name: str

@router.post("")
def _enable_plugin(body: Body):
    try:
        return enable_plugin(body.name)
    except Exception as e:
        print("error plugins:enable_plugin", e)
        raise HTTPException(status_code=500, detail=str(e))

@router.delete("/{name}")
def _disable_plugin(name: str):
    try:
        return disable_plugin(name)
    except Exception as e:
        print("error plugins:disable_plugin", e)
        raise HTTPException(status_code=500, detail=str(e))



from fastapi import APIRouter, HTTPException
from pydantic import BaseModel

from src.usecases.labels import retrieve_labels, create_label, delete_label

router = APIRouter(prefix="/labels", tags=["labels"])

@router.get("/{campaign}")
def _retrieve_labels(campaign: str):
    try:
        return retrieve_labels(campaign)
    except Exception as e:
        print("error labels:retrieve_labels", e)
        raise HTTPException(status_code=500, detail=str(e))

class Body(BaseModel):
    name: str
    campaign: str
    color: str

@router.post("")
def _create_label(body: Body):
    try:
        return create_label(body.name, body.color, body.campaign)
    except Exception as e:
        print("error labels:create_label", e)
        raise HTTPException(status_code=500, detail=str(e))

@router.delete("/{id}")
def _delete_label(id: str):
    try:
        return delete_label(id)
    except Exception as e:
        print("error labels:delete_label", e)
        raise HTTPException(status_code=500, detail=str(e))


from fastapi import APIRouter, HTTPException
from pydantic import BaseModel
from typing import Optional, List
from src.usecases.models import create_model, retrieve_models, delete_model, retrieve_one_model, inference_model, retrieve_model_label_mappings

router = APIRouter(prefix="/models", tags=["models"])

@router.get("")
def _retrieve_models(campaign: Optional[str] = None):
    try:
        return retrieve_models(campaign)
    except Exception as e:
        print("error models:retrieve_models", e)
        raise HTTPException(status_code=500, detail=str(e))
    

@router.get("/{model_id}")
def _retrieve_one_models(model_id: str):
    try:
        return retrieve_one_model(model_id)
    except Exception as e:
        print("error models:retrieve_one_model", e)
        raise HTTPException(status_code=500, detail=str(e))
    
class Body(BaseModel):
    name: str
    description: str
    url: str
    task: str
    preprocessing: Optional[List[str]] = []
    postprocessing: Optional[List[str]] = []

@router.post("")
def _create_model(body: Body):
    try:
        return create_model(body.name, body.description, body.url, body.task, body.preprocessing, body.postprocessing)
    except Exception as e:
        print("error models:create_model", e)
        raise HTTPException(status_code=500, detail=str(e))

@router.delete("/{model_id}")
def _delete_model(model_id: str):
    try:
        return delete_model(model_id)
    except Exception as e:
        print("error models:delete_model", e)
        raise HTTPException(status_code=500, detail=str(e))

class InferenceBody(BaseModel):
    image: int

@router.post("/inference/{model_id}")
def _inference_model(model_id: str, body: InferenceBody):
    try:
        return inference_model(model_id, body.image)
    except Exception as e:
        print("error models:inference_model", e)
        raise HTTPException(status_code=500, detail=str(e))
    

@router.get("/{campaign_id}/label_mappings")
def _retrieve_label_mappings(model_id: str):
    try:
        return retrieve_model_label_mappings(model_id)
    except Exception as e:
        print("error models:retrieve_model_label_mappings", e)
        raise HTTPException(status_code=500, detail=str(e))


from ...models import Label
from ...repos import LabelsDBRepo

def retrieve_labels(campaign):
    repo = LabelsDBRepo()
    data = repo.retrieve_labels(campaign)
    labels = [Label.from_tuple(d) for d in data]
    return labels


from .DBRepo import DBRepo
from datetime import datetime

class CampaignsDBRepo(DBRepo):
    def __init__(self):
        super().__init__()
        cursor = self.get_cursor()
        cursor.execute(f"""CREATE TABLE IF NOT EXISTS campaigns (
            id TEXT PRIMARY KEY,
            name TEXT NOT NULL UNIQUE,
            description TEXT NOT NULL,
            createdAt TEXT NOT NULL,
            updatedAt TEXT NOT NULL,
            image_count INTEGER NOT NULL DEFAULT 0,
            annotation_count INTEGER NOT NULL DEFAULT 0,
            eotdlDatasetId TEXT NULL
        )""")
        self.commit_and_close_db()

    def retrieve_campaigns(self):
        cursor = self.get_cursor()
        cursor.execute("SELECT id, name, description, createdAt, updatedAt, image_count, annotation_count, eotdlDatasetId FROM campaigns")
        return cursor.fetchall()
    
    def create_campaign(self, campaign):
        cursor = self.get_cursor()
        cursor.execute(f"INSERT INTO campaigns (id, name, description, createdAt, updatedAt, image_count, annotation_count, eotdlDatasetId) VALUES (?, ?, ?, ?, ?, ?, ?, ?)", (campaign.id, campaign.name, campaign.description, campaign.createdAt, campaign.updatedAt, campaign.image_count, campaign.annotation_count, campaign.eotdlDatasetId))
        self.commit_and_close_db()

    def retrieve_campaign(self, id):
        cursor = self.get_cursor()
        cursor.execute(f"SELECT * FROM campaigns WHERE id = ?", (id,))
        return cursor.fetchone()

    def delete_campaign(self, id):
        cursor = self.get_cursor()
        cursor.execute(f"DELETE FROM campaigns WHERE id = ?", (id,))
        self.commit_and_close_db()

    def retrieve_one_campaign(self, campaign_id):
        cursor = self.get_cursor()
        cursor.execute("SELECT id, name, description, createdAt, updatedAt, image_count, annotation_count, eotdlDatasetId FROM campaigns WHERE id = ?", (campaign_id,))
        return cursor.fetchone()
    
    def update_campaign(self, campaign):
        cursor = self.get_cursor()
        campaign.updatedAt = datetime.now()
        cursor.execute("UPDATE campaigns SET name = ?, description = ?, updatedAt = ?, image_count = ?, annotation_count = ?, eotdlDatasetId = ? WHERE id = ?", (campaign.name, campaign.description, campaign.updatedAt, campaign.image_count, campaign.annotation_count, campaign.eotdlDatasetId, campaign.id))
        self.commit_and_close_db()


from pydantic import BaseModel, Field
from datetime import datetime
from typing import Optional

class Plugin(BaseModel):
    id: str
    name: str
    enabled: bool = Field(default=False)
    createdAt: datetime = Field(default_factory=datetime.now)
    updatedAt: datetime = Field(default_factory=datetime.now)

    @classmethod
    def from_tuple(cls, data: tuple):
        return cls(
            id=data[0],
            name=data[1],
            enabled=data[2],
            createdAt=datetime.fromisoformat(data[3]),
            updatedAt=datetime.fromisoformat(data[4])
        )


from ...repos import PluginsDBRepo
from ...models import Plugin

def retrieve_plugins():
    repo = PluginsDBRepo()
    data = repo.retrieve_plugins()
    plugins = [Plugin.from_tuple(d) for d in data]
    return plugins


from ...repos import LabelMappingsDBRepo, LabelsDBRepo
from ...models import Label, LabelMapping

def create_label_mapping(campaing_id, model_id, label_name, index):
	labels_repo = LabelsDBRepo()
	data = labels_repo.get_label_by_name(label_name, campaing_id)
	if not data:
		raise ValueError(f"Label '{label_name}' not found in campaign '{campaing_id}'")
	label = Label.from_tuple(data)
	repo = LabelMappingsDBRepo()
	label_mapping = LabelMapping(id=repo.generate_id(), campaignId=campaing_id, modelId=model_id, labelId=label.id, output_index=index)
	if repo.label_mapping_exists(campaing_id, model_id, label.id):
		repo.update_label_mapping(label_mapping)
	else:
		repo.create_label_mapping(label_mapping)
	return label_mapping


from ...models import Campaign
from ...repos import CampaignsDBRepo

def retrieve_campaigns():
    repo = CampaignsDBRepo()
    data = repo.retrieve_campaigns()
    campaigns = [Campaign.from_tuple(d) for d in data]
    return campaigns


from fastapi import APIRouter, HTTPException, WebSocket
from pydantic import BaseModel

from src.usecases.campaigns import create_campaign, create_campaign_eotdl, retrieve_campaigns, delete_campaign, retrieve_one_campaign, create_imported_campaign, retrieve_campaign_label_mappings, update_label_mappings

router = APIRouter(prefix="/campaigns", tags=["campaigns"])

@router.get("")
def _retrieve_campaigns():
    try:
        return retrieve_campaigns()
    except Exception as e:
        print("error campaigns:get_campaigns", e)
        raise HTTPException(status_code=500, detail=str(e))
    

@router.get("/{campaign_id}")
def _retrieve_one_campaigns(campaign_id: str):
    try:
        return retrieve_one_campaign(campaign_id)
    except Exception as e:
        print("error campaigns:get_one_campaign", e)
        raise HTTPException(status_code=500, detail=str(e))
    
class Body(BaseModel):
    name: str
    description: str
    path: str

# @router.post("")
# def _create_campaign(body: Body):
#     try:
#         return create_campaign(body.name, body.description, body.path)
#     except Exception as e:
#         print("error campaigns:create_campaign", e)
#         raise HTTPException(status_code=500, detail=str(e))

@router.websocket("/create")
async def websocket_create_campaign(websocket: WebSocket):
    await websocket.accept()
    try:
        # Receive the campaign data as JSON
        data = await websocket.receive_json()
        
        # Create campaign with progress callback
        async def progress_callback(progress: float, message: str):
            await websocket.send_json({
                "progress": progress,
                "message": message,
                "status": "processing"
            })
        
        # Call create_campaign with the callback
        result = await create_campaign(
            data["name"], 
            data["description"], 
            data["path"],
            data["labels"],
            data["labelMappings"],
            progress_callback=progress_callback
        )
        
        # Send success response
        data = result.model_dump()
        await websocket.send_json({
            "status": "complete",
            "data": {
                "name": data["name"],
                "description": data["description"],
                "id": data["id"]
            }
        })
        
    except Exception as e:
        print("error campaigns:websocket_create_campaign", e)
        await websocket.send_json({
            "status": "error",
            "error": str(e)
        })
    finally:
        await websocket.close()

@router.websocket("/import")
async def websocket_create_campaign(websocket: WebSocket):
    await websocket.accept()
    try:
        # Receive the campaign data as JSON
        data = await websocket.receive_json()
        
        # Create campaign with progress callback
        async def progress_callback(progress: float, message: str):
            await websocket.send_json({
                "progress": progress,
                "message": message,
                "status": "processing"
            })
        
        # Call create_campaign with the callback
        result = await create_imported_campaign(
            data["name"], 
            data["description"], 
            data["path"],
            progress_callback=progress_callback
        )
        
        # Send success response
        data = result.model_dump()
        await websocket.send_json({
            "status": "complete",
            "data": {
                "name": data["name"],
                "description": data["description"],
                "id": data["id"]
            }
        })
        
    except Exception as e:
        print("error campaigns:websocket_import_campaign", e)
        await websocket.send_json({
            "status": "error",
            "error": str(e)
        })
    finally:
        await websocket.close()

@router.delete("/{campaign_id}")
def _delete_campaign(campaign_id: str):
    try:
        return delete_campaign(campaign_id)
    except Exception as e:
        print("error campaigns:delete_campaign", e)
        raise HTTPException(status_code=500, detail=str(e))

@router.websocket("/create-eotdl")
async def websocket_create_campaign_eotdl(websocket: WebSocket):
    await websocket.accept()
    try:
        # Receive the campaign data as JSON
        data = await websocket.receive_json()
        
        # Create campaign with progress callback
        async def progress_callback(progress: float, message: str):
            await websocket.send_json({
                "progress": progress,
                "message": message,
                "status": "processing"
            })
        
        # Call create_campaign with the callback
        result = await create_campaign_eotdl(
            data["name"], 
            data["description"], 
            data["eotdlDatasetId"],
            data["labels"],
            progress_callback=progress_callback
        )
        
        # Send success response
        data = result.model_dump()
        await websocket.send_json({
            "status": "complete",
            "data": {
                "name": data["name"],
                "description": data["description"],
                "id": data["id"]
            }
        })
        
    except Exception as e:
        print("error campaigns:websocket_create_campaign", e)
        await websocket.send_json({
            "status": "error",
            "error": str(e)
        })
    finally:
        await websocket.close()


@router.get("/{campaign_id}/label_mappings")
def _retrieve_label_mappings(campaign_id: str):
    try:
        return retrieve_campaign_label_mappings(campaign_id)
    except Exception as e:
        print("error campaigns:retrieve_campaign_label_mappings", e)
        raise HTTPException(status_code=500, detail=str(e))
    
class UpdateLabelMappingsBody(BaseModel):
    data: dict

@router.post("/{campaign_id}/label_mappings")
def _update_label_mappings(campaign_id: str, body: UpdateLabelMappingsBody):
    try:
        return update_label_mappings(campaign_id, body.data) 
    except Exception as e:
        print("error campaigns:update_label_mappings", e)
        raise HTTPException(status_code=500, detail=str(e))


from .generate_labels_json import generate_labels_json
from ..images import retrieve_images
import os
import json

def save_labels(campaign_id):
    imgs = retrieve_images(campaign_id)
    path = os.path.commonpath([image.path for image in imgs])
    labels_json = generate_labels_json(campaign_id)
    with open(os.path.join(path.split(os.sep)[0], "spai.json"), "w") as f:
        f.write(json.dumps(labels_json))


from fastapi import APIRouter, HTTPException
from pydantic import BaseModel

from src.usecases.export import save_annotations
from src.usecases.labels import save_labels
router = APIRouter(prefix="/geojson", tags=["geojson"])

class SaveBody(BaseModel):
    campaignId: str


@router.post("/save")
def _save_annotations(
    body: SaveBody,
):
    try:
        save_labels(body.campaignId)
        save_annotations(body.campaignId)
    except Exception as e:
        print("error geojson:save_geojson", e)
        return HTTPException(status_code=500, detail="Could not save geojson")


from fastapi import FastAPI
from fastapi.staticfiles import StaticFiles
from fastapi.middleware.cors import CORSMiddleware
import os

from routers import campaigns, images, models, plugins, labels, annotations, geojson

app = FastAPI(docs_url=None)

app.add_middleware(
    CORSMiddleware,
    allow_origins=["*"],
    allow_credentials=True,
    allow_methods=["*"],
    allow_headers=["*"],
)

app.include_router(campaigns.router)
app.include_router(images.router)
app.include_router(models.router)
app.include_router(plugins.router)
app.include_router(labels.router)
app.include_router(annotations.router)
app.include_router(geojson.router)

# this needs to be last in order to not override other routes
# ui is in same directory as this file
# in order for this to work with multipage apps, make sure to use trailingSlash = 'always' in svelte layout
app.mount(
    "/",
    StaticFiles(
        directory=os.path.dirname(os.path.realpath(__file__)) + "/ui", html=True
    ),
    name="ui",
)


if __name__ == "__main__":
    app()



import sqlite3
import os
import stat
import uuid

def check_permissions(path):
    st = os.stat(path)
    return bool(st.st_mode & stat.S_IWUSR)

class DBRepo:
    def __init__(self):
        home_dir = os.path.expanduser("~")
        cache_dir = os.path.join(home_dir, ".cache/scaneo/")
        if not os.path.exists(cache_dir):
            os.makedirs(cache_dir)
        if not check_permissions(cache_dir):
            print("No write permissions to cache directory")
        self.db_path = cache_dir + 'scaneo.db'
        # if os.path.exists(db_path):
        #     print("Database already exists")
        # else:
        #     print("New database created")
        

    def get_cursor(self):
        self.db = sqlite3.connect(self.db_path)
        return self.db.cursor()

    def commit_and_close_db(self):
        self.db.commit()
        self.db.close()

    def generate_id(self):
        return str(uuid.uuid4())

    # def create(self, data):
    # 	self.db.append(data)
    # 	return data



from datetime import datetime

from .DBRepo import DBRepo
from ..models import Plugin

class PluginsDBRepo(DBRepo):
    def __init__(self):
        super().__init__()
        cursor = self.get_cursor()
        cursor.execute(f"""CREATE TABLE IF NOT EXISTS plugins (
            id TEXT NOT NULL PRIMARY KEY,
            name TEXT NOT NULL UNIQUE,
            enabled BOOLEAN NOT NULL,
            createdAt TEXT NOT NULL,
            updatedAt TEXT NOT NULL
        )""")
        self.commit_and_close_db()
        self.plugins = ['eotdl']
        self.init_plugins()
        
    def init_plugins(self):
        for i, plugin in enumerate(self.plugins, start=1):
            data = self.retrieve_plugin(plugin)
            if not data:
                plugin = Plugin(id=plugin, name=plugin)
                self.add_plugin(plugin)
        # delete all plugins in table that are not in plugins
        cursor = self.get_cursor()
        cursor.execute(f"DELETE FROM plugins WHERE name NOT IN ({', '.join(['?' for _ in self.plugins])})", self.plugins)
        self.commit_and_close_db()


    def retrieve_plugins(self):
        cursor = self.get_cursor()
        cursor.execute(f"SELECT * FROM plugins")
        return cursor.fetchall()

    def retrieve_plugin(self, name):
        cursor = self.get_cursor()
        cursor.execute(f"SELECT * FROM plugins WHERE name = ?", (name,))
        return cursor.fetchone()

    def add_plugin(self, plugin):
        cursor = self.get_cursor()
        cursor.execute(f"INSERT INTO plugins (id, name, enabled, createdAt, updatedAt) VALUES (?, ?, ?, ?, ?)", (plugin.id, plugin.name, plugin.enabled, plugin.createdAt, plugin.updatedAt))
        self.commit_and_close_db()

    def update_plugin(self, plugin):
        cursor = self.get_cursor()
        plugin.updatedAt = datetime.now().isoformat()
        cursor.execute(f"UPDATE plugins SET enabled = ?, updatedAt = ? WHERE id = ?", (plugin.enabled, plugin.updatedAt, plugin.id))
        self.commit_and_close_db()


def generate_detection_geojson(bbox, label):
                min_lat, min_lng = bbox[0]
                max_lat, max_lng = bbox[1]
                
                polygon_coords = [
                    [min_lng, min_lat], [min_lng, max_lat],
                    [max_lng, max_lat], [max_lng, min_lat], 
                    [min_lng, min_lat]
                ]
                return {"type": "Feature","geometry": {"type": "Polygon","coordinates": [polygon_coords]},"properties": {"labels": [label], "bbox": bbox, "tasks": ["detection"]}}


from .create_model import create_model
from .retrieve_models import retrieve_models
from .delete_model import delete_model
from .retrieve_one_model import retrieve_one_model
from .inference_model import inference_model
from .create_label_mapping import create_label_mapping
from .retrieve_model_label_mappings import retrieve_model_label_mappings


from glob import glob
import os
import rasterio
import json

from ...models import Campaign, Image
from ...repos import CampaignsDBRepo, ImagesDBRepo
from ...errors.campaigns import CannotImportCampaign
from ...usecases.annotations import create_classification_annotation, create_detection_annotation, create_segmentation_annotation
from ...usecases.labels import create_label
def get_bbox(image):
	src = rasterio.open(image)
	bounds = src.bounds
	crs = src.crs
	if crs != "EPSG:4326":
		bounds = rasterio.warp.transform_bounds(crs, "EPSG:4326", *bounds)
	return json.dumps(bounds)

def add_annotations_to_image(image_id, annotations):
     for annotation in annotations["features"]:
        task = annotation["properties"]["tasks"][0]
        labels = annotation["properties"]["labels"][0]
        if task == "segmentation":
            layer_data = {"type": "Feature","properties": {},"geometry": annotation["geometry"]}
            create_segmentation_annotation(image_id,layer_data,labels)
        elif task == "detection":
            create_detection_annotation(image_id, annotation["properties"]["bbox"], labels)
        elif task == "classification":
            create_classification_annotation(image_id, labels)

def add_labels_to_campaign(campaign_id, labels):
    for label in labels["labels"]:
        create_label(label["name"], label["color"], campaign_id)

async def create_imported_campaign(name, description, path, progress_callback=None):	
    repo = CampaignsDBRepo()
    id = repo.generate_id()
    campaign = Campaign(id=id, name=name, description=description)
    repo.create_campaign(campaign)
    
    # Retrieve images in path recursively
    images = glob(os.path.join(path, '**', '*.tif'), recursive=True)
    geojsons = glob(os.path.join(path, '**', '*.geojson'), recursive=True)
    labels = glob(os.path.join(path, 'spai.json'))
    
    images_repo = ImagesDBRepo()
    bbs = []
    
    # Generate bounding boxes
    if images:
        for i, image in enumerate(images):
            if progress_callback is not None:
                await progress_callback((i + 1) / len(images), f"Processing {image}")
            bbs.append(get_bbox(image))
    else:
        raise ValueError("No images found in the provided path.")
    
    images_repo.create_images(images, id, bbs)

    if labels:
        with open(os.path.join(path, 'spai.json'), 'r') as f:
            add_labels_to_campaign(id, json.loads(f.read()))
        for geojson in geojsons:
            tif_path = f'{os.path.splitext(geojson)[0].replace("_labels", "")}.tif'
            if os.path.exists(tif_path):
                annotating_image = Image.from_tuple(images_repo.retrieve_image_by_path(tif_path, id))
                with open(geojson, 'r') as f:
                    add_annotations_to_image(annotating_image.id, json.loads(f.read()))
    else:
        raise CannotImportCampaign()
    return campaign


from ...repos import AnnotationsDBRepo
from ...models import Annotation

def update_detection_annotation(annotationId, bb):	
	repo = AnnotationsDBRepo()
	data = repo.retrieve_one_annotation(annotationId)
	if not data:
		raise Exception("Annotation not found")
	annotation = Annotation.from_tuple(data)
	annotation.bb = bb
	repo.update_annotation(annotation)
	return annotation



class CampaignAlreadyExistsError(Exception):
    message = "Campaign already exists"
    def __init__(self):
        super().__init__(self.message)


class CampaignDoesNotExistError(Exception):
    message = "Campaign doesn't exist"
    def __init__(self):
        super().__init__(self.message)

class CannotImportCampaign(Exception):
    message = "spai.json doesn't exist"
    def __init__(self):
        super().__init__(self.message)